#include <cuda_fp16.h>

{{corr_enum}}

extern "C" {
	/*
	  Perform corrections: NONE or OFFSET
	  Take cell_table into account when getting correction values
	  Converting to float while correcting
	  Converting to output dtype at the end
	  Shape of input data: memory cell, 1, y, x
	  Shape of offset constant: x, y, memory cell
	*/
	__global__ void correct(const {{input_data_dtype}}* data, // shape: memory cell, 1, y, x
							const unsigned short* cell_table,
	                        const unsigned char corr_flags,
							const float* offset_map,
							{{output_data_dtype}}* output) {
		const size_t X = {{pixels_x}};
		const size_t Y = {{pixels_y}};
		const size_t memory_cells = {{data_memory_cells}};
		const size_t map_memory_cells = {{constant_memory_cells}};

		const size_t memory_cell = blockIdx.x * blockDim.x + threadIdx.x;
		const size_t y = blockIdx.y * blockDim.y + threadIdx.y;
		const size_t x = blockIdx.z * blockDim.z + threadIdx.z;

		if (memory_cell >= memory_cells || y >= Y || x >= X) {
			return;
		}

		// note: strides differ from numpy strides because unit here is sizeof(...), not byte
		const size_t data_stride_x = 1;
		const size_t data_stride_y = X * data_stride_x;
		const size_t data_stride_cell = Y * data_stride_y;
		const size_t data_index = memory_cell * data_stride_cell + y * data_stride_y + x * data_stride_x;
		const float raw = (float)data[data_index];

		const size_t map_stride_x = 1;
		const size_t map_stride_y = X * map_stride_x;
		const size_t map_stride_cell = Y * map_stride_y;
		const size_t map_cell = cell_table[memory_cell];
		if (map_cell < map_memory_cells) {
			const size_t map_index = map_cell * map_stride_cell + y * map_stride_y + x * map_stride_x;
			float corrected = raw;
			if (corr_flags & OFFSET) {
				corrected -= offset_map[map_index];
			}
			{% if output_data_dtype == "half" %}
			output[data_index] = __float2half(corrected);
			{% else %}
			output[data_index] = ({{output_data_dtype}})corrected;
			{% endif %}
		} else {
			{% if output_data_dtype == "half" %}
			output[data_index] = __float2half(raw);
			{% else %}
			output[data_index] = ({{output_data_dtype}})raw;
			{% endif %}
		}
	}
}
